from __future__ import annotations
from collections.abc import Iterator
from dataclasses import dataclass
from operator import attrgetter
import os
from pathlib import Path
from traceback import format_exception
from typing import Any
import click
from click.testing import CliRunner, Result
import pytest
from pytest_mock import MockerFixture
import lineinfile
from lineinfile import (
    ALWAYS,
    CHANGED,
    AfterFirst,
    AfterLast,
    AtBOF,
    AtEOF,
    BackupWhen,
    BeforeFirst,
    BeforeLast,
    Inserter,
    add_line_to_file,
    add_line_to_string,
    ensure_terminated,
)
import lineinfile.__main__
from lineinfile.__main__ import main

CASES_DIR = Path(__file__).with_name("data") / "add_line"

INPUT = (CASES_DIR / "input.txt").read_text()


@dataclass
class AddLineCase:
    name: str
    input: str
    line: str
    args: dict[str, Any]
    options: list[str]
    output: str
    nonuniversal_lines: bool

    @property
    def changed(self) -> bool:
        return self.input != self.output


# The set of test cases has to be fetched anew for every test because the
# inserters contain mutable state.


def add_line_cases() -> Iterator[AddLineCase]:
    for cfgfile in sorted(CASES_DIR.glob("*.py")):
        cfg: dict[str, Any] = {}
        exec(cfgfile.read_text(), cfg)
        try:
            input_file = cfg["input_file"]
        except KeyError:
            source = INPUT
        else:
            with (CASES_DIR / input_file).open(newline="") as fp:
                source = fp.read()
        with cfgfile.with_suffix(".txt").open(newline="") as fp:
            output = fp.read()
        yield AddLineCase(
            name=cfgfile.with_suffix("").name,
            input=source,
            line=cfg["line"],
            args=cfg["args"],
            options=cfg["options"],
            output=output,
            nonuniversal_lines=cfg.get("nonuniversal_lines", False),
        )


def file_add_line_cases() -> Iterator[AddLineCase]:
    for c in add_line_cases():
        if not c.nonuniversal_lines:
            yield c


def listdir(dirpath: Path) -> list[str]:
    return sorted(p.name for p in dirpath.iterdir())


def show_result(r: Result) -> str:
    if r.exception is not None:
        assert isinstance(r.exc_info, tuple)
        return "".join(format_exception(*r.exc_info))
    else:
        return r.output


@pytest.mark.parametrize("case", add_line_cases(), ids=attrgetter("name"))
def test_add_line_to_string(case: AddLineCase) -> None:
    assert add_line_to_string(case.input, case.line, **case.args) == case.output


def test_backrefs_no_regexp() -> None:
    with pytest.raises(ValueError) as excinfo:
        add_line_to_string(INPUT, "gnusto=cleesh", backrefs=True)
    assert str(excinfo.value) == "backrefs=True cannot be given without regexp"


@pytest.mark.parametrize("case", file_add_line_cases(), ids=attrgetter("name"))
def test_add_line_to_file(case: AddLineCase, tmp_path: Path) -> None:
    thefile = tmp_path / "file.txt"
    thefile.write_text(case.input)
    assert add_line_to_file(thefile, case.line, **case.args) == case.changed
    assert listdir(tmp_path) == ["file.txt"]
    assert thefile.read_text() == case.output


@pytest.mark.parametrize("case", file_add_line_cases(), ids=attrgetter("name"))
def test_add_line_to_file_backup_changed(case: AddLineCase, tmp_path: Path) -> None:
    thefile = tmp_path / "file.txt"
    thefile.write_text(case.input)
    assert (
        add_line_to_file(thefile, case.line, **case.args, backup=CHANGED)
        == case.changed
    )
    if case.changed:
        assert listdir(tmp_path) == ["file.txt", "file.txt~"]
        assert thefile.with_name(thefile.name + "~").read_text() == case.input
    else:
        assert listdir(tmp_path) == ["file.txt"]
    assert thefile.read_text() == case.output


@pytest.mark.parametrize("case", file_add_line_cases(), ids=attrgetter("name"))
def test_add_line_to_file_backup_changed_custom_ext(
    case: AddLineCase, tmp_path: Path
) -> None:
    thefile = tmp_path / "file.txt"
    thefile.write_text(case.input)
    assert (
        add_line_to_file(
            thefile,
            case.line,
            **case.args,
            backup=CHANGED,
            backup_ext=".bak",
        )
        == case.changed
    )
    if case.changed:
        assert listdir(tmp_path) == ["file.txt", "file.txt.bak"]
        assert thefile.with_name(thefile.name + ".bak").read_text() == case.input
    else:
        assert listdir(tmp_path) == ["file.txt"]
    assert thefile.read_text() == case.output


@pytest.mark.parametrize("case", file_add_line_cases(), ids=attrgetter("name"))
def test_add_line_to_file_backup_always(case: AddLineCase, tmp_path: Path) -> None:
    thefile = tmp_path / "file.txt"
    thefile.write_text(case.input)
    assert (
        add_line_to_file(thefile, case.line, **case.args, backup=ALWAYS) == case.changed
    )
    assert listdir(tmp_path) == ["file.txt", "file.txt~"]
    assert thefile.with_name(thefile.name + "~").read_text() == case.input
    assert thefile.read_text() == case.output


@pytest.mark.parametrize("case", file_add_line_cases(), ids=attrgetter("name"))
def test_add_line_to_file_backup_always_custom_ext(
    case: AddLineCase, tmp_path: Path
) -> None:
    thefile = tmp_path / "file.txt"
    thefile.write_text(case.input)
    assert (
        add_line_to_file(
            thefile,
            case.line,
            **case.args,
            backup=ALWAYS,
            backup_ext=".bak",
        )
        == case.changed
    )
    assert listdir(tmp_path) == ["file.txt", "file.txt.bak"]
    assert thefile.with_name(thefile.name + ".bak").read_text() == case.input
    assert thefile.read_text() == case.output


@pytest.mark.parametrize("when", [CHANGED, ALWAYS])
def test_empty_backup_ext(when: BackupWhen) -> None:
    with pytest.raises(ValueError) as excinfo:
        add_line_to_file(
            "nonexistent.txt",
            "gnusto=cleesh",
            backup_ext="",
            backup=when,
        )
    assert str(excinfo.value) == "Cannot use empty string as backup_ext"


def test_file_line_replaces_self(tmp_path: Path) -> None:
    thefile = tmp_path / "file.txt"
    thefile.write_text(INPUT)
    assert not add_line_to_file(
        thefile,
        "bar=quux\n",
        regexp=r"^bar=",
        backup=CHANGED,
    )
    assert listdir(tmp_path) == ["file.txt"]
    assert thefile.read_text() == INPUT


@pytest.mark.parametrize("when", [CHANGED, ALWAYS])
def test_backup_file_exists(tmp_path: Path, when: BackupWhen) -> None:
    thefile = tmp_path / "file.txt"
    thefile.write_text(INPUT)
    (tmp_path / "file.txt.bak").write_text("This will be replaced.\n")
    assert add_line_to_file(
        thefile,
        "gnusto=cleesh",
        backup=when,
        backup_ext=".bak",
    )
    assert listdir(tmp_path) == ["file.txt", "file.txt.bak"]
    assert (tmp_path / "file.txt.bak").read_text() == INPUT
    assert thefile.read_text() == INPUT + "gnusto=cleesh\n"


@pytest.mark.parametrize("create", [False, True])
def test_create_file_exists(tmp_path: Path, create: bool) -> None:
    thefile = tmp_path / "file.txt"
    thefile.write_text(INPUT)
    assert add_line_to_file(thefile, "gnusto=cleesh", create=create)
    assert listdir(tmp_path) == ["file.txt"]
    assert thefile.read_text() == INPUT + "gnusto=cleesh\n"


def test_no_create_file_not_exists(tmp_path: Path) -> None:
    thefile = tmp_path / "file.txt"
    with pytest.raises(FileNotFoundError):
        add_line_to_file(thefile, "gnusto=cleesh")
    assert listdir(tmp_path) == []


def test_create_file_not_exists(tmp_path: Path) -> None:
    thefile = tmp_path / "file.txt"
    assert add_line_to_file(thefile, "gnusto=cleesh", create=True)
    assert listdir(tmp_path) == ["file.txt"]
    assert thefile.read_text() == "gnusto=cleesh\n"


@pytest.mark.parametrize("when", [CHANGED, ALWAYS])
def test_create_file_not_exists_backup(tmp_path: Path, when: BackupWhen) -> None:
    thefile = tmp_path / "file.txt"
    assert add_line_to_file(thefile, "gnusto=cleesh", create=True, backup=when)
    assert listdir(tmp_path) == ["file.txt"]
    assert thefile.read_text() == "gnusto=cleesh\n"


def test_create_file_no_change(tmp_path: Path) -> None:
    thefile = tmp_path / "file.txt"
    assert not add_line_to_file(
        thefile,
        r"\1=cleesh",
        regexp=r"^(\w+)=",
        backrefs=True,
        create=True,
    )
    assert listdir(tmp_path) == []


@pytest.mark.parametrize("when", [CHANGED, ALWAYS])
def test_backup_symlink(tmp_path: Path, when: BackupWhen) -> None:
    thefile = tmp_path / "file.txt"
    thefile.write_text(INPUT)
    linkfile = tmp_path / "link.txt"
    linkfile.symlink_to(thefile)
    assert add_line_to_file(
        linkfile,
        "gnusto=cleesh",
        backup=when,
        backup_ext=".bak",
    )
    assert listdir(tmp_path) == ["file.txt", "link.txt", "link.txt.bak"]
    assert linkfile.is_symlink()
    assert not (tmp_path / "link.txt.bak").is_symlink()
    assert (tmp_path / "link.txt.bak").read_text() == INPUT
    assert thefile.read_text() == INPUT + "gnusto=cleesh\n"


def test_backup_symlink_no_change(tmp_path: Path) -> None:
    thefile = tmp_path / "file.txt"
    thefile.write_text(INPUT)
    linkfile = tmp_path / "link.txt"
    linkfile.symlink_to(thefile)
    assert not add_line_to_file(
        linkfile,
        "foo=apple",
        backup=ALWAYS,
        backup_ext=".bak",
    )
    assert listdir(tmp_path) == ["file.txt", "link.txt", "link.txt.bak"]
    assert linkfile.is_symlink()
    assert not (tmp_path / "link.txt.bak").is_symlink()
    assert (tmp_path / "link.txt.bak").read_text() == INPUT
    assert thefile.read_text() == INPUT


STRING_DEFAULTS = {
    "regexp": None,
    "backrefs": False,
    "match_first": False,
    "inserter": None,
}

CLI_DEFAULTS = {
    **STRING_DEFAULTS,
    "backup": None,
    "backup_ext": None,
    "create": False,
    "encoding": "utf-8",
}


@pytest.mark.parametrize("case", file_add_line_cases(), ids=attrgetter("name"))
@pytest.mark.parametrize(
    "backup_opts,backup_args",
    [
        ([], {}),
        (["--backup"], {"backup": CHANGED}),
        (["--backup-changed"], {"backup": CHANGED}),
        (["--backup", "-i.bak"], {"backup": CHANGED, "backup_ext": ".bak"}),
        (["-i.bak"], {"backup": CHANGED, "backup_ext": ".bak"}),
        (["--backup-ext=.bak"], {"backup": CHANGED, "backup_ext": ".bak"}),
        (["--backup-always"], {"backup": ALWAYS}),
        (["--backup-always", "-i.bak"], {"backup": ALWAYS, "backup_ext": ".bak"}),
        (["-i.bak", "--backup-always"], {"backup": ALWAYS, "backup_ext": ".bak"}),
        (["--backup-changed", "--backup-always"], {"backup": ALWAYS}),
        (["--backup-always", "--backup-changed"], {"backup": CHANGED}),
        (["--create"], {"create": True}),
    ],
)
def test_cli_add(
    case: AddLineCase,
    backup_opts: list[str],
    backup_args: dict[str, Any],
    mocker: MockerFixture,
) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("file.txt").touch()
        add_line_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_file",
            return_value=case.changed,
        )
        r = runner.invoke(
            main,
            ["add"] + case.options + backup_opts + [case.line, "file.txt"],
            standalone_mode=False,
        )
    assert r.exit_code == 0, show_result(r)
    assert r.output == ""
    args = {**CLI_DEFAULTS, **case.args, **backup_args}
    if args["regexp"] is not None and not isinstance(args["regexp"], str):
        args["regexp"] = args["regexp"].pattern
    add_line_mock.assert_called_once_with("file.txt", case.line, **args)


@pytest.mark.parametrize(
    "backup_opts,backup_args",
    [
        ([], {}),
        (["--create"], {"create": True}),
    ],
)
def test_cli_add_file_not_exists_no_error(
    backup_opts: list[str], backup_args: dict[str, Any], mocker: MockerFixture
) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        add_line_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_file",
            return_value=True,
        )
        r = runner.invoke(
            main,
            ["add"] + backup_opts + ["gnusto=cleesh", "file.txt"],
            standalone_mode=False,
        )
    assert r.exit_code == 0, show_result(r)
    assert r.output == ""
    args = {**CLI_DEFAULTS, **backup_args}
    add_line_mock.assert_called_once_with("file.txt", "gnusto=cleesh", **args)


@pytest.mark.parametrize(
    "opts,inserter",
    [
        ([], None),
        (["--after-first", "foo"], AfterFirst("foo")),
        (["--after-first", "foo", "--before-last", "bar"], BeforeLast("bar")),
        (["-B", "bar", "-a", "foo"], AfterFirst("foo")),
        (["--bof"], AtBOF()),
        (["--bof", "-A", "foo"], AfterLast("foo")),
        (["-A", "foo", "--bof"], AtBOF()),
        (["--bof", "--eof"], AtEOF()),
        (["--eof", "--bof"], AtBOF()),
        (["--bof", "--eof", "-b", "foo"], BeforeFirst("foo")),
        (["-a", "foo", "-A", "bar", "-b", "baz", "-B", "quux"], BeforeLast("quux")),
        (
            ["-a", "foo", "-A", "bar", "--bof", "-b", "baz", "-B", "quux"],
            BeforeLast("quux"),
        ),
        (["-a", "foo", "-a", "bar"], AfterFirst("bar")),
        pytest.param(
            ["-a", "foo", "-b", "quux", "-a", "bar"],
            AfterFirst("bar"),
            marks=pytest.mark.xfail(reason="Click doesn't work that way."),
        ),
    ],
)
def test_cli_add_inserter_options(
    opts: list[str], inserter: Inserter | None, mocker: MockerFixture
) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("file.txt").touch()
        add_line_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_file",
            return_value=True,
        )
        r = runner.invoke(
            main,
            ["add"] + opts + ["gnusto=cleesh", "file.txt"],
            standalone_mode=False,
        )
    assert r.exit_code == 0, show_result(r)
    assert r.output == ""
    args = {**CLI_DEFAULTS, "inserter": inserter}
    add_line_mock.assert_called_once_with("file.txt", "gnusto=cleesh", **args)


@pytest.mark.parametrize(
    "opts,match_first",
    [
        ([], False),
        (["--match-first"], True),
        (["--match-last"], False),
        (["-m", "-M"], False),
        (["-M", "-m"], True),
        (["-m", "-M", "-m"], True),
        (["-M", "-m", "-M"], False),
    ],
)
def test_cli_add_match_options(
    opts: list[str], match_first: bool, mocker: MockerFixture
) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("file.txt").touch()
        add_line_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_file",
            return_value=True,
        )
        r = runner.invoke(
            main,
            ["add"] + opts + ["gnusto=cleesh", "file.txt"],
            standalone_mode=False,
        )
    assert r.exit_code == 0, show_result(r)
    assert r.output == ""
    args = {**CLI_DEFAULTS, "match_first": match_first}
    add_line_mock.assert_called_once_with("file.txt", "gnusto=cleesh", **args)


def test_cli_add_backrefs_no_regex(mocker: MockerFixture) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("file.txt").touch()
        add_line_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_file",
            return_value=True,
        )
        r = runner.invoke(
            main,
            ["add", "--backrefs", "gnusto=cleesh", "file.txt"],
            standalone_mode=False,
        )
    assert r.exit_code != 0
    assert isinstance(r.exception, click.UsageError)
    assert str(r.exception) == "--backrefs cannot be specified without --regexp"
    add_line_mock.assert_not_called()


def test_cli_add_empty_backup_ext(mocker: MockerFixture) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("file.txt").touch()
        add_line_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_file",
            return_value=True,
        )
        r = runner.invoke(
            main,
            ["add", "--backup-ext=", "gnusto=cleesh", "file.txt"],
            standalone_mode=False,
        )
    assert r.exit_code != 0
    assert isinstance(r.exception, click.UsageError)
    assert str(r.exception) == "--backup-ext cannot be empty"
    add_line_mock.assert_not_called()


@pytest.mark.parametrize("case", file_add_line_cases(), ids=attrgetter("name"))
@pytest.mark.parametrize("input_args", [[], ["-"]])
def test_cli_add_stdin(
    case: AddLineCase, input_args: list[str], mocker: MockerFixture
) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("-").touch()
        add_line_file_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_file",
            return_value=case.changed,
        )
        add_line_str_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_string",
            return_value=case.output,
        )
        r = runner.invoke(
            main,
            ["add"] + case.options + [case.line] + input_args,
            input=case.input,
            standalone_mode=False,
        )
    assert r.exit_code == 0, show_result(r)
    assert r.output == case.output
    args = {**STRING_DEFAULTS, **case.args}
    if args["regexp"] is not None and not isinstance(args["regexp"], str):
        args["regexp"] = args["regexp"].pattern
    add_line_file_mock.assert_not_called()
    add_line_str_mock.assert_called_once_with(case.input, case.line, **args)


@pytest.mark.parametrize("input_args", [[], ["-"]])
@pytest.mark.parametrize(
    "file_arg,err_arg",
    [
        ("--backup", "--backup-changed"),
        ("--backup-changed", "--backup-changed"),
        ("--backup-always", "--backup-always"),
        ("-i.bak", "--backup-ext"),
        ("--create", "--create"),
    ],
)
def test_cli_add_stdin_bad_file_args(
    file_arg: str, err_arg: str, input_args: list[str], mocker: MockerFixture
) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("file.txt").touch()
        add_line_file_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_file",
        )
        add_line_str_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_string",
        )
        r = runner.invoke(
            main,
            ["add", file_arg, "gnusto=cleesh"] + input_args,
            input="This is test text.\n",
            standalone_mode=False,
        )
    assert r.exit_code != 0
    assert isinstance(r.exception, click.UsageError)
    assert str(r.exception) == (
        f"{err_arg} cannot be set when reading from standard input."
    )
    add_line_file_mock.assert_not_called()
    add_line_str_mock.assert_not_called()


@pytest.mark.parametrize("case", file_add_line_cases(), ids=attrgetter("name"))
def test_cli_add_outfile(case: AddLineCase, mocker: MockerFixture) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        thefile = Path("file.txt")
        thefile.write_text(case.input)
        add_line_file_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_file",
        )
        add_line_str_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_string",
            return_value=case.output,
        )
        r = runner.invoke(
            main,
            ["add"] + case.options + ["--outfile=out.txt", case.line, "file.txt"],
            standalone_mode=False,
        )
        assert r.exit_code == 0, show_result(r)
        assert r.output == ""
        assert sorted(os.listdir()) == ["file.txt", "out.txt"]
        assert thefile.read_text() == case.input
        assert Path("out.txt").read_text() == case.output
    args = {**STRING_DEFAULTS, **case.args}
    if args["regexp"] is not None and not isinstance(args["regexp"], str):
        args["regexp"] = args["regexp"].pattern
    add_line_file_mock.assert_not_called()
    add_line_str_mock.assert_called_once_with(case.input, case.line, **args)


@pytest.mark.parametrize("case", file_add_line_cases(), ids=attrgetter("name"))
@pytest.mark.parametrize("input_args", [[], ["-"]])
def test_cli_add_stdin_outfile(
    case: AddLineCase, input_args: list[str], mocker: MockerFixture
) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("-").touch()
        add_line_file_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_file",
            return_value=case.changed,
        )
        add_line_str_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_string",
            return_value=case.output,
        )
        r = runner.invoke(
            main,
            ["add"] + case.options + ["-oout.txt", case.line] + input_args,
            input=case.input,
            standalone_mode=False,
        )
        assert r.exit_code == 0, show_result(r)
        assert r.output == ""
        assert sorted(os.listdir()) == ["-", "out.txt"]
        assert Path("-").read_text() == ""
        assert Path("out.txt").read_text() == case.output
    args = {**STRING_DEFAULTS, **case.args}
    if args["regexp"] is not None and not isinstance(args["regexp"], str):
        args["regexp"] = args["regexp"].pattern
    add_line_file_mock.assert_not_called()
    add_line_str_mock.assert_called_once_with(case.input, case.line, **args)


@pytest.mark.parametrize("case", file_add_line_cases(), ids=attrgetter("name"))
def test_cli_add_outfile_stdout(case: AddLineCase, mocker: MockerFixture) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        thefile = Path("file.txt")
        thefile.write_text(case.input)
        add_line_file_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_file",
        )
        add_line_str_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_string",
            return_value=case.output,
        )
        r = runner.invoke(
            main,
            ["add"] + case.options + ["--outfile", "-", case.line, "file.txt"],
            standalone_mode=False,
        )
        assert r.exit_code == 0, show_result(r)
        assert r.output == case.output
        assert os.listdir() == ["file.txt"]
        assert thefile.read_text() == case.input
    args = {**STRING_DEFAULTS, **case.args}
    if args["regexp"] is not None and not isinstance(args["regexp"], str):
        args["regexp"] = args["regexp"].pattern
    add_line_file_mock.assert_not_called()
    add_line_str_mock.assert_called_once_with(case.input, case.line, **args)


@pytest.mark.parametrize(
    "file_arg,err_arg",
    [
        ("--backup", "--backup-changed"),
        ("--backup-changed", "--backup-changed"),
        ("--backup-always", "--backup-always"),
        ("-i.bak", "--backup-ext"),
        ("--create", "--create"),
    ],
)
def test_cli_add_outfile_bad_file_args(
    file_arg: str, err_arg: str, mocker: MockerFixture
) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("file.txt").touch()
        add_line_file_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_file",
        )
        add_line_str_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_string",
        )
        r = runner.invoke(
            main,
            ["add", "-o", "out.txt", file_arg, "gnusto=cleesh", "file.txt"],
            standalone_mode=False,
        )
    assert r.exit_code != 0
    assert isinstance(r.exception, click.UsageError)
    assert str(r.exception) == f"{err_arg} is incompatible with --outfile."
    add_line_file_mock.assert_not_called()
    add_line_str_mock.assert_not_called()


@pytest.mark.parametrize("case", file_add_line_cases(), ids=attrgetter("name"))
def test_cli_add_outfile_is_infile(case: AddLineCase, mocker: MockerFixture) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        thefile = Path("file.txt")
        thefile.write_text(case.input)
        add_line_file_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_file",
        )
        add_line_str_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_string",
            return_value=case.output,
        )
        r = runner.invoke(
            main,
            ["add"] + case.options + ["--outfile=file.txt", case.line, "file.txt"],
            standalone_mode=False,
        )
        assert r.exit_code == 0, show_result(r)
        assert r.output == ""
        assert os.listdir() == ["file.txt"]
        assert thefile.read_text() == case.output
    args = {**STRING_DEFAULTS, **case.args}
    if args["regexp"] is not None and not isinstance(args["regexp"], str):
        args["regexp"] = args["regexp"].pattern
    add_line_file_mock.assert_not_called()
    add_line_str_mock.assert_called_once_with(case.input, case.line, **args)


@pytest.mark.parametrize(
    "escline,line",
    [
        ("foo", "foo"),
        (r"foo\n", "foo\n"),
        (r"foo\\n", "foo\\n"),
        (r"foo\\\n", "foo\\\n"),
        (r"foo\012", "foo\n"),
        (r"foo\x0A", "foo\n"),
        (r"foo\u000A", "foo\n"),
        (r"foo\\bar", r"foo\bar"),
        (r"foo\'bar", "foo'bar"),
        (r"foo\"bar", 'foo"bar'),
        (r"foo\abar", "foo\abar"),
        (r"foo\bbar", "foo\bbar"),
        (r"foo\fbar", "foo\fbar"),
        (r"foo\tbar", "foo\tbar"),
        (r"foo\vbar", "foo\vbar"),
        (r"\U0001F410", "\U0001F410"),
        ("åéîøü", "åéîøü"),
        (r"\u2603", "\u2603"),
        ("\u2603", "\u2603"),
        ("\U0001F410", "\U0001F410"),
        (r"\N{SNOWMAN}", "\u2603"),
    ],
)
def test_cli_add_backslashed(escline: str, line: str, mocker: MockerFixture) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        thefile = Path("file.txt")
        thefile.write_text(INPUT)
        add_line_file_spy = mocker.spy(lineinfile.__main__, "add_line_to_file")
        r = runner.invoke(
            main,
            ["add", escline, "file.txt"],
            standalone_mode=False,
        )
        assert r.exit_code == 0, show_result(r)
        assert r.output == ""
        assert os.listdir() == ["file.txt"]
        assert thefile.read_text() == INPUT + ensure_terminated(line)
    add_line_file_spy.assert_called_once_with("file.txt", line, **CLI_DEFAULTS)


@pytest.mark.parametrize(
    "escline,line",
    [
        ("foo", "foo"),
        (r"foo\n", "foo\n"),
        (r"foo\\n", "foo\\n"),
        (r"foo\\\n", "foo\\\n"),
        (r"foo\012", "foo\n"),
        pytest.param(
            r"foo\x0A",
            "foo\n",
            marks=pytest.mark.xfail(reason="Not supported by Match.expand()"),
        ),
        pytest.param(
            r"foo\u000A",
            "foo\n",
            marks=pytest.mark.xfail(reason="Not supported by Match.expand()"),
        ),
        (r"foo\\bar", r"foo\bar"),
        pytest.param(
            r"foo\'bar",
            "foo'bar",
            marks=pytest.mark.xfail(reason="Not supported by Match.expand()"),
        ),
        pytest.param(
            r"foo\"bar",
            'foo"bar',
            marks=pytest.mark.xfail(reason="Not supported by Match.expand()"),
        ),
        (r"foo\abar", "foo\abar"),
        (r"foo\bbar", "foo\bbar"),
        (r"foo\fbar", "foo\fbar"),
        (r"foo\tbar", "foo\tbar"),
        (r"foo\vbar", "foo\vbar"),
        pytest.param(
            r"\U0001F410",
            "\U0001F410",
            marks=pytest.mark.xfail(reason="Not supported by Match.expand()"),
        ),
        ("åéîøü", "åéîøü"),
        pytest.param(
            r"\u2603",
            "\u2603",
            marks=pytest.mark.xfail(reason="Not supported by Match.expand()"),
        ),
        ("\u2603", "\u2603"),
        ("\U0001F410", "\U0001F410"),
        pytest.param(
            r"\N{SNOWMAN}",
            "\u2603",
            marks=pytest.mark.xfail(reason="Not supported by Match.expand()"),
        ),
    ],
)
def test_cli_add_backslashed_backrefs(
    escline: str, line: str, mocker: MockerFixture
) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        thefile = Path("file.txt")
        thefile.write_text(INPUT + "replaceme\n")
        add_line_file_spy = mocker.spy(lineinfile.__main__, "add_line_to_file")
        r = runner.invoke(
            main,
            ["add", "--backrefs", "-e", "replaceme", escline, "file.txt"],
            standalone_mode=False,
        )
        assert r.exit_code == 0, show_result(r)
        assert r.output == ""
        assert os.listdir() == ["file.txt"]
        assert thefile.read_text() == INPUT + ensure_terminated(line)
    args = {**CLI_DEFAULTS, "regexp": "replaceme", "backrefs": True}
    add_line_file_spy.assert_called_once_with("file.txt", escline, **args)


@pytest.mark.parametrize(
    "line",
    [
        "gnusto=cleesh",
        "-ecleesh",
        "--gnusto=cleesh",
    ],
)
def test_cli_add_line_opt_file(line: str, mocker: MockerFixture) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("file.txt").touch()
        add_line_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_file",
            return_value=True,
        )
        r = runner.invoke(
            main,
            ["add", "-L", line, "file.txt"],
            standalone_mode=False,
        )
    assert r.exit_code == 0, show_result(r)
    assert r.output == ""
    add_line_mock.assert_called_once_with("file.txt", line, **CLI_DEFAULTS)


@pytest.mark.parametrize(
    "line",
    [
        "gnusto=cleesh",
        "-ecleesh",
        "--gnusto=cleesh",
    ],
)
@pytest.mark.parametrize("input_args", [[], ["-"]])
def test_cli_add_line_opt_stdin(
    input_args: list[str], line: str, mocker: MockerFixture
) -> None:
    runner = CliRunner()
    output = INPUT + line + "\n"
    with runner.isolated_filesystem():
        Path("file.txt").touch()
        add_line_file_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_file",
            return_value=True,
        )
        add_line_str_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_string",
            return_value=output,
        )
        r = runner.invoke(
            main,
            ["add", "-L", line] + input_args,
            input=INPUT,
            standalone_mode=False,
        )
    assert r.exit_code == 0, show_result(r)
    assert r.output == output
    add_line_file_mock.assert_not_called()
    add_line_str_mock.assert_called_once_with(INPUT, line, **STRING_DEFAULTS)


def test_cli_add_line_opt_two_args(mocker: MockerFixture) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("file.txt").touch()
        add_line_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_file",
            return_value=True,
        )
        r = runner.invoke(
            main,
            ["add", "--line", "gnusto=cleesh", "foo=bar", "file.txt"],
            standalone_mode=False,
        )
    assert r.exit_code != 0
    assert isinstance(r.exception, click.UsageError)
    assert str(r.exception) == "-L/--line given with too many positional arguments"
    add_line_mock.assert_not_called()


def test_cli_add_no_line_opt_no_args(mocker: MockerFixture) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        add_line_mock = mocker.patch(
            "lineinfile.__main__.add_line_to_file",
            return_value=True,
        )
        r = runner.invoke(main, ["add"], standalone_mode=False)
    assert r.exit_code != 0
    assert isinstance(r.exception, click.UsageError)
    assert str(r.exception) == "No LINE given"
    add_line_mock.assert_not_called()


def test_add_line_to_file_encoding(mocker: MockerFixture, tmp_path: Path) -> None:
    thefile = tmp_path / "file.txt"
    thefile.write_text(INPUT, encoding="utf-16")
    add_line_str_spy = mocker.spy(lineinfile, "add_line_to_string")
    assert add_line_to_file(thefile, "gnusto=cleesh", encoding="utf-16")
    add_line_str_spy.assert_called_once_with(
        INPUT,
        "gnusto=cleesh",
        **STRING_DEFAULTS,
    )
    assert listdir(tmp_path) == ["file.txt"]
    assert thefile.read_text(encoding="utf-16") == INPUT + "gnusto=cleesh\n"


def test_add_line_to_file_encoding_errors(
    mocker: MockerFixture, tmp_path: Path
) -> None:
    thefile = tmp_path / "file.txt"
    thefile.write_text(
        "edh=\xF0\na-tilde-degrees=\xC3\xB0\n",
        encoding="latin-1",
    )
    add_line_str_spy = mocker.spy(lineinfile, "add_line_to_string")
    assert add_line_to_file(
        thefile,
        "edh=\xF0",
        encoding="utf-8",
        errors="surrogateescape",
    )
    add_line_str_spy.assert_called_once_with(
        "edh=\uDCF0\na-tilde-degrees=\xF0\n",
        "edh=\xF0",
        **STRING_DEFAULTS,
    )
    assert listdir(tmp_path) == ["file.txt"]
    assert thefile.read_text(encoding="latin-1") == (
        "edh=\xF0\na-tilde-degrees=\xC3\xB0\nedh=\xC3\xB0\n"
    )


def test_add_line_to_file_encoding_errors_backup(
    mocker: MockerFixture, tmp_path: Path
) -> None:
    thefile = tmp_path / "file.txt"
    thefile.write_text(
        "edh=\xF0\na-tilde-degrees=\xC3\xB0\n",
        encoding="latin-1",
    )
    add_line_str_spy = mocker.spy(lineinfile, "add_line_to_string")
    assert add_line_to_file(
        thefile,
        "edh=\xF0",
        encoding="utf-8",
        errors="surrogateescape",
        backup=CHANGED,
    )
    add_line_str_spy.assert_called_once_with(
        "edh=\uDCF0\na-tilde-degrees=\xF0\n",
        "edh=\xF0",
        **STRING_DEFAULTS,
    )
    assert listdir(tmp_path) == ["file.txt", "file.txt~"]
    assert thefile.read_text(encoding="latin-1") == (
        "edh=\xF0\na-tilde-degrees=\xC3\xB0\nedh=\xC3\xB0\n"
    )
    assert thefile.with_name(thefile.name + "~").read_text(encoding="latin-1") == (
        "edh=\xF0\na-tilde-degrees=\xC3\xB0\n"
    )


def test_after_first_reusable() -> None:
    inserter = AfterFirst("^foo=")
    assert (
        add_line_to_string(
            "foo=bar\nbar=baz\nbaz=quux\n",
            "gnusto=cleesh",
            inserter=inserter,
        )
        == "foo=bar\ngnusto=cleesh\nbar=baz\nbaz=quux\n"
    )
    assert (
        add_line_to_string(
            "food=yummy\nfoo=icky\nfo=misspelled\n",
            "gnusto=cleesh",
            inserter=inserter,
        )
        == "food=yummy\nfoo=icky\ngnusto=cleesh\nfo=misspelled\n"
    )



"""
Add & remove lines in files by regex

Inspired by (but not affiliated with) `the Ansible module of the same name`__,
``lineinfile`` provides a command and library for adding a line to a file if
it's not already there and for removing lines matching a pattern from a file.
There are options for using a regex to find a line to update or to determine
which line to insert before or after.  There are options for backing up the
modified file with a custom file extension and for treating a nonexistent file
as though it's just empty.  There's even an option for determining the line to
insert based on capturing groups in the matching regex.

__ https://docs.ansible.com/ansible/latest/collections/ansible/builtin/
   lineinfile_module.html

Unlike the Ansible module, this package does not perform any management of file
attributes; those must be set externally.

Visit <https://github.com/jwodder/lineinfile> for more information.
"""

from __future__ import annotations
from abc import ABC, abstractmethod
from enum import Enum
import os
from pathlib import Path
import re
from shutil import copystat
from typing import Any, Optional, Pattern, Union

__version__ = "0.4.1"
__author__ = "John Thorvald Wodder II"
__author_email__ = "lineinfile@varonathe.org"
__license__ = "MIT"
__url__ = "https://github.com/jwodder/lineinfile"

__all__ = [
    "ALWAYS",
    "AfterFirst",
    "AfterLast",
    "AtBOF",
    "AtEOF",
    "BackupWhen",
    "BeforeFirst",
    "BeforeLast",
    "CHANGED",
    "add_line_to_file",
    "add_line_to_string",
    "remove_lines_from_file",
    "remove_lines_from_string",
]

Patternish = Union[str, Pattern[str]]


class Inserter(ABC):
    @abstractmethod
    def update_state(
        self, state: Optional[int], lineno: int, line: str
    ) -> Optional[int]: ...

    def get_feeder(self) -> LineFeeder:
        return LineFeeder(self)


class LineFeeder:
    def __init__(self, inserter: Inserter):
        self.inserter = inserter
        self.state: Optional[int] = None

    def feed(self, lineno: int, line: str) -> None:
        self.state = self.inserter.update_state(self.state, lineno, line)

    def get_index(self) -> Optional[int]:
        return self.state


class AtBOF(Inserter):
    """Inserter that always inserts at the beginning of the file"""

    def update_state(
        self, _state: Optional[int], _lineno: int, _line: str
    ) -> Optional[int]:
        return 0

    def __eq__(self, other: Any) -> bool:
        if type(self) is type(other):
            return True
        else:
            return NotImplemented


class AtEOF(Inserter):
    """Inserter that always inserts at the end of the file"""

    def update_state(
        self, _state: Optional[int], _lineno: int, _line: str
    ) -> Optional[int]:
        return None

    def __eq__(self, other: Any) -> bool:
        if type(self) is type(other):
            return True
        else:
            return NotImplemented


class PatternInserter(Inserter):
    def __init__(self, pattern: Patternish) -> None:
        self.pattern: re.Pattern[str] = re.compile(pattern)

    def __eq__(self, other: Any) -> bool:
        if type(self) is type(other):
            return bool(self.pattern == other.pattern)
        else:
            return NotImplemented

    def __repr__(self) -> str:
        return "{0.__module__}.{0.__name__}(pattern={1.pattern!r})".format(
            type(self), self
        )


class AfterFirst(PatternInserter):
    """
    Inserter that inserts after the first input line that matches the given
    regular expression (either a string or a compiled pattern object), or at
    the end of the file if no line matches
    """

    def update_state(
        self, state: Optional[int], lineno: int, line: str
    ) -> Optional[int]:
        if state is None and self.pattern.search(line):
            return lineno + 1
        else:
            return state


class AfterLast(PatternInserter):
    """
    Inserter that inserts after the last input line that matches the given
    regular expression (either a string or a compiled pattern object), or at
    the end of the file if no line matches
    """

    def update_state(
        self, state: Optional[int], lineno: int, line: str
    ) -> Optional[int]:
        if self.pattern.search(line):
            return lineno + 1
        else:
            return state


class BeforeFirst(PatternInserter):
    """
    Inserter that inserts before the first input line that matches the given
    regular expression (either a string or a compiled pattern object), or at
    the end of the file if no line matches
    """

    def update_state(
        self, state: Optional[int], lineno: int, line: str
    ) -> Optional[int]:
        if state is None and self.pattern.search(line):
            return lineno
        else:
            return state


class BeforeLast(PatternInserter):
    """
    Inserter that inserts before the last input line that matches the given
    regular expression (either a string or a compiled pattern object), or at
    the end of the file if no line matches
    """

    def update_state(
        self, state: Optional[int], lineno: int, line: str
    ) -> Optional[int]:
        if self.pattern.search(line):
            return lineno
        else:
            return state


class Matcher(ABC):
    def __init__(self, pattern: Patternish) -> None:
        self.pattern: re.Pattern[str] = re.compile(pattern)
        self.i: Optional[int] = None
        self.m: Optional[re.Match[str]] = None

    @abstractmethod
    def feed(self, i: int, line: str) -> None: ...

    def get_index(self) -> Optional[int]:
        return self.i

    def expand(self, line: str) -> str:
        if self.m is None:
            raise ValueError("No match to expand")  # pragma: no cover
        else:
            return self.m.expand(line)


class MatchFirst(Matcher):
    def feed(self, i: int, line: str) -> None:
        if self.i is None:
            m = self.pattern.search(line)
            if m:
                self.i = i
                self.m = m


class MatchLast(Matcher):
    def feed(self, i: int, line: str) -> None:
        m = self.pattern.search(line)
        if m:
            self.i = i
            self.m = m


class ExactMatchFirst:
    def __init__(self, line: str) -> None:
        self.line: str = chomp(line)
        self.i: Optional[int] = None

    def feed(self, i: int, line: str) -> None:
        if self.i is None and self.line == chomp(line):
            self.i = i

    def get_index(self) -> Optional[int]:
        return self.i


class ExactMatchLast:
    def __init__(self, line: str) -> None:
        self.line: str = chomp(line)
        self.i: Optional[int] = None

    def feed(self, i: int, line: str) -> None:
        if self.line == chomp(line):
            self.i = i

    def get_index(self) -> Optional[int]:
        return self.i


class BackupWhen(Enum):
    CHANGED = "CHANGED"
    ALWAYS = "ALWAYS"


CHANGED = BackupWhen.CHANGED
ALWAYS = BackupWhen.ALWAYS


def add_line_to_string(
    s: str,
    line: str,
    regexp: Optional[Patternish] = None,
    inserter: Optional[Inserter] = None,
    match_first: bool = False,
    backrefs: bool = False,
) -> str:
    """
    Add the given ``line`` to the string ``s`` if it is not already present and
    return the result.  If ``regexp`` is set to a regular expression (either a
    string or a compiled pattern object) and it matches any lines in the input,
    ``line`` will replace the last matching line (or the first matching line,
    if ``match_first=True``).  If the regular expression does not match any
    lines (or no regular expression is specified) and ``line`` is not found in
    the input, the line is inserted at the end of the input by default; this
    can be changed by passing the appropriate object as the ``inserter``
    argument; see "Inserters_" below.

    When ``backrefs`` is true, if ``regexp`` matches, the capturing groups in
    the regular expression are used to expand any ``\\n``, ``\\g<n>``, or
    ``\\g<name>`` backreferences in ``line``, and the resulting string replaces
    the matched line in the input.  If ``backrefs`` is true and ``regexp`` does
    not match, the input is left unchanged.  It is an error to set ``backrefs``
    to true without also setting ``regexp``.
    """
    line_matcher: ExactMatchFirst | ExactMatchLast
    if match_first:
        line_matcher = ExactMatchFirst(line)
    else:
        line_matcher = ExactMatchLast(line)
    rgx: Optional[Matcher]
    if regexp is None:
        rgx = None
        if backrefs:
            raise ValueError("backrefs=True cannot be given without regexp")
    elif match_first:
        rgx = MatchFirst(regexp)
    else:
        rgx = MatchLast(regexp)
    insfeeder = (AtEOF() if inserter is None else inserter).get_feeder()
    lines = ascii_splitlines(s)
    for i, ln in enumerate(lines):
        line_matcher.feed(i, ln)
        if rgx is not None:
            rgx.feed(i, ln)
        insfeeder.feed(i, ln)
    match_point = None if rgx is None else rgx.get_index()
    if match_point is None:
        if backrefs:
            return s
        match_point = line_matcher.get_index()
    if match_point is not None:
        assert match_point < len(lines)
        if backrefs:
            assert rgx is not None and rgx.get_index() is not None
            line = rgx.expand(line)
        lines[match_point] = ensure_terminated(line)
    else:
        insert_point = insfeeder.get_index()
        if insert_point is None:
            if lines:
                lines[-1] = ensure_terminated(lines[-1])
            lines.append(ensure_terminated(line))
        else:
            if lines and insert_point == len(lines):
                lines[-1] = ensure_terminated(lines[-1])
            lines.insert(insert_point, ensure_terminated(line))
    return "".join(lines)


def add_line_to_file(
    filepath: str | bytes | os.PathLike[str] | os.PathLike[bytes],
    line: str,
    regexp: Optional[Patternish] = None,
    inserter: Optional[Inserter] = None,
    match_first: bool = False,
    backrefs: bool = False,
    backup: Optional[BackupWhen] = None,
    backup_ext: Optional[str] = None,
    create: bool = False,
    encoding: Optional[str] = None,
    errors: Optional[str] = None,
) -> bool:
    """
    Add the given ``line`` to the file at ``filepath`` if it is not already
    present.  Returns ``True`` if the file is modified.  If ``regexp`` is set
    to a regular expression (either a string or a compiled pattern object) and
    it matches any lines in the file, ``line`` will replace the last matching
    line (or the first matching line, if ``match_first=True``).  If the regular
    expression does not match any lines (or no regular expression is specified)
    and ``line`` is not found in the file, the line is inserted at the end of
    the file by default; this can be changed by passing the appropriate object
    as the ``inserter`` argument; see "Inserters_" below.

    When ``backrefs`` is true, if ``regexp`` matches, the capturing groups in
    the regular expression are used to expand any ``\\n``, ``\\g<n>``, or
    ``\\g<name>`` backreferences in ``line``, and the resulting string replaces
    the matched line in the input.  If ``backrefs`` is true and ``regexp`` does
    not match, the file is left unchanged.  It is an error to set ``backrefs``
    to true without also setting ``regexp``.

    When ``backup`` is set to ``lineinfile.CHANGED``, a backup of the file's
    original contents is created if the file is modified.  When ``backup`` is
    set to ``lineinfile.ALWAYS``, a backup is always created, regardless of
    whether the file is modified.  The name of the backup file will be the same
    as the original, with the value of ``backup_ext`` (default: ``~``)
    appended.

    If ``create`` is true and ``filepath`` does not exist, pretend it's empty
    instead of erroring, and create it with the results of the operation.  No
    backup file will ever be created for a nonexistent file.  If ``filepath``
    does not exist and no changes are made (because ``backrefs`` was set and
    ``regexp`` didn't match), the file will not be created.
    """
    bext = "~" if backup_ext is None else backup_ext
    if backup is not None and not bext:
        raise ValueError("Cannot use empty string as backup_ext")
    p = Path(os.fsdecode(filepath))
    try:
        before = p.read_text(encoding=encoding, errors=errors)
    except FileNotFoundError:
        if create:
            before = ""
            creating = True
        else:
            raise
    else:
        creating = False
    after = add_line_to_string(
        before,
        line,
        regexp=regexp,
        inserter=inserter,
        match_first=match_first,
        backrefs=backrefs,
    )
    if not creating and backup is not None and (after != before or backup is ALWAYS):
        bak = p.with_name(p.name + bext)
        bak.write_text(before, encoding=encoding, errors=errors)
        copystat(p, bak)
    if after != before:
        p.write_text(after, encoding=encoding, errors=errors)
        return True
    else:
        return False


def remove_lines_from_string(s: str, regexp: Patternish) -> str:
    """
    Delete all lines from the string ``s`` that match the regular expression
    ``regexp`` (either a string or a compiled pattern object) and return the
    result.
    """
    lines = ascii_splitlines(s)
    lines = [ln for ln in lines if not re.search(regexp, ln)]
    return "".join(lines)


def remove_lines_from_file(
    filepath: str | bytes | os.PathLike[str] | os.PathLike[bytes],
    regexp: Patternish,
    backup: Optional[BackupWhen] = None,
    backup_ext: Optional[str] = None,
    encoding: Optional[str] = None,
    errors: Optional[str] = None,
) -> bool:
    """
    Delete all lines from the file at ``filepath`` that match the regular
    expression ``regexp`` (either a string or a compiled pattern object).
    Returns ``True`` if the file is modified.

    When ``backup`` is set to ``lineinfile.CHANGED``, a backup of the file's
    original contents is created if the file is modified.  When ``backup`` is
    set to ``lineinfile.ALWAYS``, a backup is always created, regardless of
    whether the file is modified.  The name of the backup file will be the same
    as the original, with the value of ``backup_ext`` (default: ``~``)
    appended.
    """
    bext = "~" if backup_ext is None else backup_ext
    if backup is not None and not bext:
        raise ValueError("Cannot use empty string as backup_ext")
    p = Path(os.fsdecode(filepath))
    before = p.read_text(encoding=encoding, errors=errors)
    after = remove_lines_from_string(before, regexp)
    if backup is not None and (after != before or backup is ALWAYS):
        bak = p.with_name(p.name + bext)
        bak.write_text(before, encoding=encoding, errors=errors)
        copystat(p, bak)
    if after != before:
        p.write_text(after, encoding=encoding, errors=errors)
        return True
    else:
        return False


def ensure_terminated(s: str) -> str:
    if s.endswith(("\r\n", "\n", "\r")):
        return s
    else:
        return s + "\n"


EOL_RGX = re.compile(r"\r\n?|\n")


def ascii_splitlines(s: str) -> list[str]:
    """
    Like `str.splitlines(True)`, except it only treats LF, CR LF, and CR as
    line endings
    """
    lines = []
    lastend = 0
    for m in EOL_RGX.finditer(s):
        lines.append(s[lastend : m.end()])
        lastend = m.end()
    if lastend < len(s):
        lines.append(s[lastend:])
    return lines


def chomp(s: str) -> str:
    """Remove a LF, CR, or CR LF line ending from a string"""
    if s.endswith("\n"):
        s = s[:-1]
    if s.endswith("\r"):
        s = s[:-1]
    return s


def unescape(s: str) -> str:
    # <https://stackoverflow.com/a/57192592/744178>
    return s.encode("latin-1", "backslashreplace").decode("unicode_escape")



from __future__ import annotations
import pytest
from lineinfile import ascii_splitlines, chomp, ensure_terminated, unescape


@pytest.mark.parametrize(
    "s,lines",
    [
        ("", []),
        ("foobar", ["foobar"]),
        ("foo\n", ["foo\n"]),
        ("foo\r", ["foo\r"]),
        ("foo\r\n", ["foo\r\n"]),
        ("foo\n\r", ["foo\n", "\r"]),
        ("foo\nbar", ["foo\n", "bar"]),
        ("foo\rbar", ["foo\r", "bar"]),
        ("foo\r\nbar", ["foo\r\n", "bar"]),
        ("foo\n\rbar", ["foo\n", "\r", "bar"]),
        ("foo\n\nbar", ["foo\n", "\n", "bar"]),
        ("foo\n\nbar\n", ["foo\n", "\n", "bar\n"]),
        (
            "Why\vare\fthere\x1Cso\x1Ddang\x1Emany\x85line\u2028separator\u2029"
            "characters?",
            [
                "Why\vare\fthere\x1Cso\x1Ddang\x1Emany\x85line\u2028separator\u2029"
                "characters?"
            ],
        ),
    ],
)
def test_ascii_splitlines(s: str, lines: list[str]) -> None:
    assert ascii_splitlines(s) == lines


@pytest.mark.parametrize(
    "s,chomped",
    [
        ("foobar", "foobar"),
        ("foobar\n", "foobar"),
        ("foobar\r\n", "foobar"),
        ("foobar\r", "foobar"),
        ("foobar\n\r", "foobar\n"),
        ("foobar\n\n", "foobar\n"),
        ("foobar\nbaz", "foobar\nbaz"),
    ],
)
def test_chomp(s: str, chomped: str) -> None:
    assert chomp(s) == chomped


@pytest.mark.parametrize(
    "s,terminated",
    [
        ("foobar", "foobar\n"),
        ("foobar\n", "foobar\n"),
        ("foobar\r", "foobar\r"),
        ("foobar\r\n", "foobar\r\n"),
        ("foobar\n\r", "foobar\n\r"),
        ("foo\nbar", "foo\nbar\n"),
    ],
)
def test_ensure_terminated(s: str, terminated: str) -> None:
    assert ensure_terminated(s) == terminated


@pytest.mark.parametrize(
    "src,dest",
    [
        ("foo", "foo"),
        (r"foo\n", "foo\n"),
        (r"foo\\n", "foo\\n"),
        (r"foo\\\n", "foo\\\n"),
        (r"foo\012", "foo\n"),
        (r"foo\x0A", "foo\n"),
        (r"foo\u000A", "foo\n"),
        (r"foo\\bar", r"foo\bar"),
        (r"foo\'bar", "foo'bar"),
        (r"foo\"bar", 'foo"bar'),
        (r"foo\abar", "foo\abar"),
        (r"foo\bbar", "foo\bbar"),
        (r"foo\fbar", "foo\fbar"),
        (r"foo\rbar", "foo\rbar"),
        (r"foo\tbar", "foo\tbar"),
        (r"foo\vbar", "foo\vbar"),
        (r"\U0001F410", "\U0001F410"),
        ("åéîøü", "åéîøü"),
        (r"\u2603", "\u2603"),
        ("\u2603", "\u2603"),
        ("\U0001F410", "\U0001F410"),
        (r"\N{SNOWMAN}", "\u2603"),
    ],
)
def test_unescape(src: str, dest: str) -> None:
    assert unescape(src) == dest



from lineinfile import AfterFirst

line = r"gnusto=\1:\2"
args = {
    "regexp": r"^(notinfile)=(\w+)",
    "backrefs": True,
    "inserter": AfterFirst(r"^foo="),
}
options = ["-e", r"^(notinfile)=(\w+)", "--backrefs", "-a", "^foo="]



from __future__ import annotations
from collections.abc import Iterator
from dataclasses import dataclass
from operator import attrgetter
import os
from pathlib import Path
from traceback import format_exception
from typing import Any
import click
from click.testing import CliRunner, Result
import pytest
from pytest_mock import MockerFixture
import lineinfile
from lineinfile import (
    ALWAYS,
    CHANGED,
    BackupWhen,
    remove_lines_from_file,
    remove_lines_from_string,
)
from lineinfile.__main__ import main

CASES_DIR = Path(__file__).with_name("data") / "remove_lines"

INPUT = (CASES_DIR / "input.txt").read_text()


@dataclass
class RemoveLinesCase:
    name: str
    input: str
    regexp: str
    output: str

    @property
    def changed(self) -> bool:
        return self.input != self.output


def remove_lines_cases() -> Iterator[RemoveLinesCase]:
    for cfgfile in sorted(CASES_DIR.glob("*.py")):
        cfg: dict[str, Any] = {}
        exec(cfgfile.read_text(), cfg)
        try:
            input_file = cfg["input_file"]
        except KeyError:
            source = INPUT
        else:
            with (CASES_DIR / input_file).open(newline="") as fp:
                source = fp.read()
        with cfgfile.with_suffix(".txt").open(newline="") as fp:
            output = fp.read()
        yield RemoveLinesCase(
            name=cfgfile.with_suffix("").name,
            input=source,
            regexp=cfg["regexp"],
            output=output,
        )


def listdir(dirpath: Path) -> list[str]:
    return sorted(p.name for p in dirpath.iterdir())


def show_result(r: Result) -> str:
    if r.exception is not None:
        assert isinstance(r.exc_info, tuple)
        return "".join(format_exception(*r.exc_info))
    else:
        return r.output


REMOVE_LINES_CASES = list(remove_lines_cases())

CHANGE_CASE = next(c for c in REMOVE_LINES_CASES if c.changed)
NO_CHANGE_CASE = next(c for c in REMOVE_LINES_CASES if not c.changed)


@pytest.mark.parametrize("case", REMOVE_LINES_CASES, ids=attrgetter("name"))
def test_remove_lines_from_string(case: RemoveLinesCase) -> None:
    assert remove_lines_from_string(case.input, case.regexp) == case.output


@pytest.mark.parametrize("case", REMOVE_LINES_CASES, ids=attrgetter("name"))
def test_remove_lines_from_file(case: RemoveLinesCase, tmp_path: Path) -> None:
    thefile = tmp_path / "file.txt"
    thefile.write_text(case.input)
    assert remove_lines_from_file(thefile, case.regexp) == case.changed
    assert listdir(tmp_path) == ["file.txt"]
    assert thefile.read_text() == case.output


@pytest.mark.parametrize("case", REMOVE_LINES_CASES, ids=attrgetter("name"))
def test_remove_lines_from_file_backup_changed(
    case: RemoveLinesCase, tmp_path: Path
) -> None:
    thefile = tmp_path / "file.txt"
    thefile.write_text(case.input)
    assert remove_lines_from_file(thefile, case.regexp, backup=CHANGED) == case.changed
    if case.changed:
        assert listdir(tmp_path) == ["file.txt", "file.txt~"]
        assert thefile.with_name(thefile.name + "~").read_text() == case.input
    else:
        assert listdir(tmp_path) == ["file.txt"]
    assert thefile.read_text() == case.output


@pytest.mark.parametrize("case", REMOVE_LINES_CASES, ids=attrgetter("name"))
def test_remove_lines_from_file_backup_changed_custom_ext(
    case: RemoveLinesCase, tmp_path: Path
) -> None:
    thefile = tmp_path / "file.txt"
    thefile.write_text(case.input)
    assert (
        remove_lines_from_file(
            thefile,
            case.regexp,
            backup=CHANGED,
            backup_ext=".bak",
        )
        == case.changed
    )
    if case.changed:
        assert listdir(tmp_path) == ["file.txt", "file.txt.bak"]
        assert thefile.with_name(thefile.name + ".bak").read_text() == case.input
    else:
        assert listdir(tmp_path) == ["file.txt"]
    assert thefile.read_text() == case.output


@pytest.mark.parametrize("case", REMOVE_LINES_CASES, ids=attrgetter("name"))
def test_remove_lines_from_file_backup_always(
    case: RemoveLinesCase, tmp_path: Path
) -> None:
    thefile = tmp_path / "file.txt"
    thefile.write_text(case.input)
    assert remove_lines_from_file(thefile, case.regexp, backup=ALWAYS) == case.changed
    assert listdir(tmp_path) == ["file.txt", "file.txt~"]
    assert thefile.with_name(thefile.name + "~").read_text() == case.input
    assert thefile.read_text() == case.output


@pytest.mark.parametrize("case", REMOVE_LINES_CASES, ids=attrgetter("name"))
def test_remove_lines_from_file_backup_always_custom_ext(
    case: RemoveLinesCase, tmp_path: Path
) -> None:
    thefile = tmp_path / "file.txt"
    thefile.write_text(case.input)
    assert (
        remove_lines_from_file(
            thefile,
            case.regexp,
            backup=ALWAYS,
            backup_ext=".bak",
        )
        == case.changed
    )
    assert listdir(tmp_path) == ["file.txt", "file.txt.bak"]
    assert thefile.with_name(thefile.name + ".bak").read_text() == case.input
    assert thefile.read_text() == case.output


@pytest.mark.parametrize("when", [CHANGED, ALWAYS])
def test_empty_backup_ext(when: BackupWhen) -> None:
    with pytest.raises(ValueError) as excinfo:
        remove_lines_from_file(
            "nonexistent.txt",
            r"^gnusto=",
            backup_ext="",
            backup=when,
        )
    assert str(excinfo.value) == "Cannot use empty string as backup_ext"


@pytest.mark.parametrize("when", [CHANGED, ALWAYS])
def test_backup_file_exists(tmp_path: Path, when: BackupWhen) -> None:
    thefile = tmp_path / "file.txt"
    thefile.write_text(CHANGE_CASE.input)
    (tmp_path / "file.txt.bak").write_text("This will be replaced.\n")
    assert remove_lines_from_file(
        thefile,
        CHANGE_CASE.regexp,
        backup=when,
        backup_ext=".bak",
    )
    assert listdir(tmp_path) == ["file.txt", "file.txt.bak"]
    assert (tmp_path / "file.txt.bak").read_text() == CHANGE_CASE.input
    assert thefile.read_text() == CHANGE_CASE.output


@pytest.mark.parametrize("when", [CHANGED, ALWAYS])
def test_backup_symlink(tmp_path: Path, when: BackupWhen) -> None:
    thefile = tmp_path / "file.txt"
    thefile.write_text(CHANGE_CASE.input)
    linkfile = tmp_path / "link.txt"
    linkfile.symlink_to(thefile)
    assert remove_lines_from_file(
        linkfile,
        CHANGE_CASE.regexp,
        backup=when,
        backup_ext=".bak",
    )
    assert listdir(tmp_path) == ["file.txt", "link.txt", "link.txt.bak"]
    assert linkfile.is_symlink()
    assert not (tmp_path / "link.txt.bak").is_symlink()
    assert (tmp_path / "link.txt.bak").read_text() == CHANGE_CASE.input
    assert thefile.read_text() == CHANGE_CASE.output


def test_backup_symlink_no_change(tmp_path: Path) -> None:
    thefile = tmp_path / "file.txt"
    thefile.write_text(NO_CHANGE_CASE.input)
    linkfile = tmp_path / "link.txt"
    linkfile.symlink_to(thefile)
    assert not remove_lines_from_file(
        linkfile,
        NO_CHANGE_CASE.regexp,
        backup=ALWAYS,
        backup_ext=".bak",
    )
    assert listdir(tmp_path) == ["file.txt", "link.txt", "link.txt.bak"]
    assert linkfile.is_symlink()
    assert not (tmp_path / "link.txt.bak").is_symlink()
    assert (tmp_path / "link.txt.bak").read_text() == NO_CHANGE_CASE.input
    assert thefile.read_text() == NO_CHANGE_CASE.output


CLI_DEFAULTS = {
    "backup": None,
    "backup_ext": None,
    "encoding": "utf-8",
}


@pytest.mark.parametrize(
    "opts,args",
    [
        ([], {}),
        (["--backup"], {"backup": CHANGED}),
        (["--backup-changed"], {"backup": CHANGED}),
        (["--backup", "-i.bak"], {"backup": CHANGED, "backup_ext": ".bak"}),
        (["-i.bak"], {"backup": CHANGED, "backup_ext": ".bak"}),
        (["--backup-ext=.bak"], {"backup": CHANGED, "backup_ext": ".bak"}),
        (["--backup-always"], {"backup": ALWAYS}),
        (["--backup-always", "-i.bak"], {"backup": ALWAYS, "backup_ext": ".bak"}),
        (["-i.bak", "--backup-always"], {"backup": ALWAYS, "backup_ext": ".bak"}),
        (["--backup-changed", "--backup-always"], {"backup": ALWAYS}),
        (["--backup-always", "--backup-changed"], {"backup": CHANGED}),
    ],
)
def test_cli_remove(
    opts: list[str], args: dict[str, Any], mocker: MockerFixture
) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("file.txt").touch()
        remove_lines_mock = mocker.patch(
            "lineinfile.__main__.remove_lines_from_file",
            return_value=True,
        )
        r = runner.invoke(
            main,
            ["remove"] + opts + ["^foo=", "file.txt"],
            standalone_mode=False,
        )
    assert r.exit_code == 0, show_result(r)
    assert r.output == ""
    fargs = {**CLI_DEFAULTS, **args}
    remove_lines_mock.assert_called_once_with("file.txt", "^foo=", **fargs)


def test_cli_remove_empty_backup_ext(mocker: MockerFixture) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("file.txt").touch()
        remove_lines_mock = mocker.patch(
            "lineinfile.__main__.remove_lines_from_file",
            return_value=True,
        )
        r = runner.invoke(
            main,
            ["remove", "--backup-ext=", "^foo=", "file.txt"],
            standalone_mode=False,
        )
    assert r.exit_code != 0
    assert isinstance(r.exception, click.UsageError)
    assert str(r.exception) == "--backup-ext cannot be empty"
    remove_lines_mock.assert_not_called()


@pytest.mark.parametrize("input_args", [[], ["-"]])
def test_cli_remove_stdin(input_args: list[str], mocker: MockerFixture) -> None:
    runner = CliRunner()
    output = remove_lines_from_string(INPUT, "^foo=")
    with runner.isolated_filesystem():
        Path("-").touch()
        remove_lines_file_mock = mocker.patch(
            "lineinfile.__main__.remove_lines_from_file",
            return_value=True,
        )
        remove_lines_str_mock = mocker.patch(
            "lineinfile.__main__.remove_lines_from_string",
            return_value=output,
        )
        r = runner.invoke(
            main,
            ["remove", "^foo="] + input_args,
            input=INPUT,
            standalone_mode=False,
        )
    assert r.exit_code == 0, show_result(r)
    assert r.output == output
    remove_lines_file_mock.assert_not_called()
    remove_lines_str_mock.assert_called_once_with(INPUT, "^foo=")


@pytest.mark.parametrize("input_args", [[], ["-"]])
@pytest.mark.parametrize(
    "file_arg,err_arg",
    [
        ("--backup", "--backup-changed"),
        ("--backup-changed", "--backup-changed"),
        ("--backup-always", "--backup-always"),
        ("-i.bak", "--backup-ext"),
    ],
)
def test_cli_remove_stdin_bad_file_args(
    file_arg: str, err_arg: str, input_args: list[str], mocker: MockerFixture
) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("file.txt").touch()
        remove_lines_file_mock = mocker.patch(
            "lineinfile.__main__.remove_lines_from_file",
        )
        remove_lines_str_mock = mocker.patch(
            "lineinfile.__main__.remove_lines_from_string",
        )
        r = runner.invoke(
            main,
            ["remove", file_arg, "^gnusto="] + input_args,
            input="This is test text.\n",
            standalone_mode=False,
        )
    assert r.exit_code != 0
    assert isinstance(r.exception, click.UsageError)
    assert str(r.exception) == (
        f"{err_arg} cannot be set when reading from standard input."
    )
    remove_lines_file_mock.assert_not_called()
    remove_lines_str_mock.assert_not_called()


def test_cli_remove_outfile(mocker: MockerFixture) -> None:
    runner = CliRunner()
    output = remove_lines_from_string(INPUT, "^foo=")
    with runner.isolated_filesystem():
        thefile = Path("file.txt")
        thefile.write_text(INPUT)
        remove_lines_file_mock = mocker.patch(
            "lineinfile.__main__.remove_lines_from_file",
        )
        remove_lines_str_mock = mocker.patch(
            "lineinfile.__main__.remove_lines_from_string",
            return_value=output,
        )
        r = runner.invoke(
            main,
            ["remove", "--outfile=out.txt", "^foo=", "file.txt"],
            standalone_mode=False,
        )
        assert r.exit_code == 0, show_result(r)
        assert r.output == ""
        assert sorted(os.listdir()) == ["file.txt", "out.txt"]
        assert thefile.read_text() == INPUT
        assert Path("out.txt").read_text() == output
    remove_lines_file_mock.assert_not_called()
    remove_lines_str_mock.assert_called_once_with(INPUT, "^foo=")


@pytest.mark.parametrize("input_args", [[], ["-"]])
def test_cli_remove_stdin_outfile(input_args: list[str], mocker: MockerFixture) -> None:
    runner = CliRunner()
    output = remove_lines_from_string(INPUT, "^foo=")
    with runner.isolated_filesystem():
        Path("-").touch()
        remove_lines_file_mock = mocker.patch(
            "lineinfile.__main__.remove_lines_from_file",
        )
        remove_lines_str_mock = mocker.patch(
            "lineinfile.__main__.remove_lines_from_string",
            return_value=output,
        )
        r = runner.invoke(
            main,
            ["remove", "-oout.txt", "^foo="] + input_args,
            input=INPUT,
            standalone_mode=False,
        )
        assert r.exit_code == 0, show_result(r)
        assert r.output == ""
        assert sorted(os.listdir()) == ["-", "out.txt"]
        assert Path("-").read_text() == ""
        assert Path("out.txt").read_text() == output
    remove_lines_file_mock.assert_not_called()
    remove_lines_str_mock.assert_called_once_with(INPUT, "^foo=")


def test_cli_remove_outfile_stdout(mocker: MockerFixture) -> None:
    runner = CliRunner()
    output = remove_lines_from_string(INPUT, "^foo=")
    with runner.isolated_filesystem():
        thefile = Path("file.txt")
        thefile.write_text(INPUT)
        remove_lines_file_mock = mocker.patch(
            "lineinfile.__main__.remove_lines_from_file",
        )
        remove_lines_str_mock = mocker.patch(
            "lineinfile.__main__.remove_lines_from_string",
            return_value=output,
        )
        r = runner.invoke(
            main,
            ["remove", "--outfile", "-", "^foo=", "file.txt"],
            standalone_mode=False,
        )
        assert r.exit_code == 0, show_result(r)
        assert r.output == output
        assert os.listdir() == ["file.txt"]
        assert thefile.read_text() == INPUT
    remove_lines_file_mock.assert_not_called()
    remove_lines_str_mock.assert_called_once_with(INPUT, "^foo=")


@pytest.mark.parametrize(
    "file_arg,err_arg",
    [
        ("--backup", "--backup-changed"),
        ("--backup-changed", "--backup-changed"),
        ("--backup-always", "--backup-always"),
        ("-i.bak", "--backup-ext"),
    ],
)
def test_cli_remove_outfile_bad_file_args(
    file_arg: str, err_arg: str, mocker: MockerFixture
) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("file.txt").touch()
        remove_lines_file_mock = mocker.patch(
            "lineinfile.__main__.remove_lines_from_file",
        )
        remove_lines_str_mock = mocker.patch(
            "lineinfile.__main__.remove_lines_from_string",
        )
        r = runner.invoke(
            main,
            ["remove", "-o", "out.txt", file_arg, "^foo=", "file.txt"],
            standalone_mode=False,
        )
    assert r.exit_code != 0
    assert isinstance(r.exception, click.UsageError)
    assert str(r.exception) == f"{err_arg} is incompatible with --outfile."
    remove_lines_file_mock.assert_not_called()
    remove_lines_str_mock.assert_not_called()


def test_cli_remove_outfile_is_infile(mocker: MockerFixture) -> None:
    runner = CliRunner()
    output = remove_lines_from_string(INPUT, "^foo=")
    with runner.isolated_filesystem():
        thefile = Path("file.txt")
        thefile.write_text(INPUT)
        remove_lines_file_mock = mocker.patch(
            "lineinfile.__main__.remove_lines_from_file",
        )
        remove_lines_str_mock = mocker.patch(
            "lineinfile.__main__.remove_lines_from_string",
            return_value=output,
        )
        r = runner.invoke(
            main,
            ["remove", "--outfile=file.txt", "^foo=", "file.txt"],
            standalone_mode=False,
        )
        assert r.exit_code == 0, show_result(r)
        assert r.output == ""
        assert os.listdir() == ["file.txt"]
        assert thefile.read_text() == output
    remove_lines_file_mock.assert_not_called()
    remove_lines_str_mock.assert_called_once_with(INPUT, "^foo=")


@pytest.mark.parametrize("regexp", ["^foo=", "-Lfoo=", "--foo=bar"])
def test_cli_remove_regexp_opt_file(regexp: str, mocker: MockerFixture) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("file.txt").touch()
        remove_lines_mock = mocker.patch(
            "lineinfile.__main__.remove_lines_from_file",
            return_value=True,
        )
        r = runner.invoke(
            main,
            ["remove", "-e", regexp, "file.txt"],
            standalone_mode=False,
        )
    assert r.exit_code == 0, show_result(r)
    assert r.output == ""
    remove_lines_mock.assert_called_once_with("file.txt", regexp, **CLI_DEFAULTS)


@pytest.mark.parametrize("regexp", ["^foo=", "-Lfoo=", "--foo=bar"])
@pytest.mark.parametrize("input_args", [[], ["-"]])
def test_cli_remove_regexp_opt_stdin(
    input_args: list[str], regexp: str, mocker: MockerFixture
) -> None:
    runner = CliRunner()
    output = remove_lines_from_string(INPUT, regexp)
    with runner.isolated_filesystem():
        Path("file.txt").touch()
        remove_lines_file_mock = mocker.patch(
            "lineinfile.__main__.remove_lines_from_file",
            return_value=INPUT != output,
        )
        remove_lines_str_mock = mocker.patch(
            "lineinfile.__main__.remove_lines_from_string",
            return_value=output,
        )
        r = runner.invoke(
            main,
            ["remove", "-e", regexp] + input_args,
            input=INPUT,
            standalone_mode=False,
        )
    assert r.exit_code == 0, show_result(r)
    assert r.output == output
    remove_lines_file_mock.assert_not_called()
    remove_lines_str_mock.assert_called_once_with(INPUT, regexp)


def test_cli_remove_regexp_opt_two_args(mocker: MockerFixture) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        Path("file.txt").touch()
        remove_lines_mock = mocker.patch(
            "lineinfile.__main__.remove_lines_from_file",
            return_value=True,
        )
        r = runner.invoke(
            main,
            ["remove", "--regexp", "^gnusto=", "bar$", "file.txt"],
            standalone_mode=False,
        )
    assert r.exit_code != 0
    assert isinstance(r.exception, click.UsageError)
    assert str(r.exception) == "-e/--regexp given with too many positional arguments"
    remove_lines_mock.assert_not_called()


def test_cli_remove_no_regexp_opt_no_args(mocker: MockerFixture) -> None:
    runner = CliRunner()
    with runner.isolated_filesystem():
        remove_lines_mock = mocker.patch(
            "lineinfile.__main__.remove_lines_from_file",
            return_value=True,
        )
        r = runner.invoke(main, ["remove"], standalone_mode=False)
    assert r.exit_code != 0
    assert isinstance(r.exception, click.UsageError)
    assert str(r.exception) == "No REGEXP given"
    remove_lines_mock.assert_not_called()


def test_remove_lines_from_file_encoding(mocker: MockerFixture, tmp_path: Path) -> None:
    thefile = tmp_path / "file.txt"
    thefile.write_text(INPUT, encoding="utf-16")
    remove_lines_str_spy = mocker.spy(lineinfile, "remove_lines_from_string")
    assert remove_lines_from_file(thefile, "^foo", encoding="utf-16")
    remove_lines_str_spy.assert_called_once_with(INPUT, "^foo")
    assert listdir(tmp_path) == ["file.txt"]
    assert thefile.read_text(encoding="utf-16") == remove_lines_from_string(
        INPUT, "^foo"
    )


def test_remove_lines_from_file_encoding_errors(
    mocker: MockerFixture, tmp_path: Path
) -> None:
    thefile = tmp_path / "file.txt"
    thefile.write_text(
        "edh=\xF0\na-tilde-degrees=\xC3\xB0\n",
        encoding="latin-1",
    )
    remove_lines_str_spy = mocker.spy(lineinfile, "remove_lines_from_string")
    assert remove_lines_from_file(
        thefile,
        "\uDCF0",
        encoding="utf-8",
        errors="surrogateescape",
    )
    remove_lines_str_spy.assert_called_once_with(
        "edh=\uDCF0\na-tilde-degrees=\xF0\n",
        "\uDCF0",
    )
    assert listdir(tmp_path) == ["file.txt"]
    assert thefile.read_text(encoding="latin-1") == "a-tilde-degrees=\xC3\xB0\n"


def test_remove_lines_from_file_encoding_errors_backup(
    mocker: MockerFixture, tmp_path: Path
) -> None:
    thefile = tmp_path / "file.txt"
    thefile.write_text(
        "edh=\xF0\na-tilde-degrees=\xC3\xB0\n",
        encoding="latin-1",
    )
    remove_lines_str_spy = mocker.spy(lineinfile, "remove_lines_from_string")
    assert remove_lines_from_file(
        thefile,
        "\xF0",
        encoding="utf-8",
        errors="surrogateescape",
        backup=CHANGED,
    )
    remove_lines_str_spy.assert_called_once_with(
        "edh=\uDCF0\na-tilde-degrees=\xF0\n",
        "\xF0",
    )
    assert listdir(tmp_path) == ["file.txt", "file.txt~"]
    assert thefile.read_text(encoding="latin-1") == "edh=\xF0\n"
    assert thefile.with_name(thefile.name + "~").read_text(encoding="latin-1") == (
        "edh=\xF0\na-tilde-degrees=\xC3\xB0\n"
    )
